import os, sys

import environments

# import python packages
import time
from copy import deepcopy

# import third-party packages
import numpy as np
import torch
import torch.optim as optim
from multiprocessing import Process, Queue, Event
import pickle

# import our packages
from scalarization_methods import WeightedSumScalarization
from sample import Sample
from task import Task
from ep import EP
from population_2d import Population as Population2d
from population_3d import Population as Population3d
from opt_graph import OptGraph
from utils import generate_weights_batch_dfs, print_info
from warm_up import initialize_warm_up_batch
from mopg import MOPG_worker, Extension_worker
from plot import plot
import wandb

from pymoo.factory import get_performance_indicator
#from pymoo.util.nds.non_dominated_sorting import NonDominatedSorting
from pymoo.indicators.hv import Hypervolume
from pymoo.util.nds.non_dominated_sorting import NonDominatedSorting

def eval(eval_sample, ref):
    r = eval_sample
    perf_ind = get_performance_indicator("hv", ref_point = -ref)
    hv = perf_ind.do(-r)
    print(hv)
    return hv

def run(args):
    # wandb
    print(args.save_dir)
    os.makedirs(os.path.join(args.save_dir, 'objs'), exist_ok = True)
    mode = 'disabled'
    wandb.init(
        project="CMORL",
        group="hopper-v3",
        name="hopper-v3",
        mode=mode
    )

    # --------------------> Preparation <-------------------- #
    np.random.seed(args.seed)
    torch.manual_seed(args.seed)
    torch.set_default_dtype(torch.float64)
    torch.set_num_threads(1)
    device = torch.device("cpu")
    os.environ["PYTHONHASHSEED"] = str(args.seed)

    start_time = time.time()

    # initialize ep and population and opt_graph
    print(args.ref_point)
    ref = np.array(args.ref_point)
    print(ref)
    ep = EP(ref, args.num_select)
    
    opt_graph = OptGraph()
    
    # Construct tasks for warm up
    elite_batch, scalarization_batch = initialize_warm_up_batch(args, device)

    for sample, scalarization in zip(elite_batch, scalarization_batch):
        sample.optgraph_id = opt_graph.insert(deepcopy(scalarization.weights), deepcopy(sample.objs), -1)
    
    print_info('\n------------------------------- Initialization Stage -------------------------------')    
        
    offspring_batch = np.array([])

    # --------------------> RL Optimization <-------------------- #
    # compose task for each elite
    iteration = 0
    task_batch = []
    for elite, scalarization in \
            zip(elite_batch, scalarization_batch):
        task_batch.append(Task(elite, scalarization)) # each task is a (policy, weight) pair
        
    # Compute update steps for initialization
    num_update = int(args.num_init_steps // len(task_batch)// args.num_steps)

    # run MOPG for each task in parallel
    processes = []
    results_queue = Queue()
    done_event = Event()
        
    for task_id, task in enumerate(task_batch):
        p = Process(target = MOPG_worker, \
            args = (args, task_id, task, device, iteration, num_update, start_time, results_queue, done_event))
        p.start()
        processes.append(p)
                                                        
    # collect MOPG results for offsprings and insert objs into objs buffer
    all_offspring_batch = [[] for _ in range(len(processes))]
    cnt_done_workers = 0
    while cnt_done_workers < len(processes):
        rl_results = results_queue.get()
        task_id, offsprings = rl_results['task_id'], rl_results['offspring_batch']
        for sample in offsprings:
            all_offspring_batch[task_id].append(Sample.copy_from(sample))
        if rl_results['done']:
            cnt_done_workers += 1
        
    # put all intermidiate policies into all_sample_batch for EP update
    all_sample_batch = [] 
    # store the last policy for each optimization weight for RA
    last_offspring_batch = [None] * len(processes) 
    # only the policies with iteration % update_iter = 0 are inserted into offspring_batch for population update
    # after warm-up stage, it's equivalent to the last_offspring_batch
    offspring_batch = [] 
    for task_id in range(len(processes)):
        offsprings = all_offspring_batch[task_id]
        prev_node_id = task_batch[task_id].sample.optgraph_id
        opt_weights = deepcopy(task_batch[task_id].scalarization.weights).detach().numpy()
        for i, sample in enumerate(offsprings):
            all_sample_batch.append(sample)
            if (i + 1) % args.update_iter == 0:
                prev_node_id = opt_graph.insert(opt_weights, deepcopy(sample.objs), prev_node_id)
                sample.optgraph_id = prev_node_id
                offspring_batch.append(sample)
        last_offspring_batch[task_id] = offsprings[-1]

    done_event.set()

    # -----------------------> Update EP <----------------------- #
    # update EP and population
    ep.update(all_sample_batch)

    # ------------------- > Task Selection <--------------------- #
    elite_batch = ep.sample_batch
        
    print_info('Selected Tasks:')
    for i in range(len(elite_batch)):
        print_info('objs = {}'.format(elite_batch[i].objs))
        
    # save
    os.makedirs(os.path.join(args.save_dir, 'objs'), exist_ok = True)
    os.makedirs(os.path.join(args.save_dir, 'ep_objs'), exist_ok = True)
    reward_list = []
    for sample in all_sample_batch:
        reward_list.append(sample.objs)
    reward_list_array = np.array(reward_list)
    torch.save(reward_list_array,os.path.join(args.save_dir, 'objs', 'init.pt'))
    hv = eval(reward_list_array, ref)
    print(hv)
    #hv = plot(reward_list_array, 0, args.save_dir)
    #wandb.log({"hv": hv})

    print_info('\n------------------------------- Extension Stage -------------------------------') 
    
    # Compute update steps for initialization
    num_update = int((args.num_time_steps - args.num_init_steps) // args.num_select // args.num_steps // 2 // args.obj_num)

    for extension_iter in range (2):   
        extension_start_time = time.time()

        # run MOPG for each task in parallel
        processes = []
        results_queue = Queue()
        done_event = Event()
            
        for sample_id, sample in enumerate(elite_batch):
            p = Process(target = Extension_worker, \
                args = (args, sample_id, sample, device, iteration, num_update, extension_start_time, results_queue, done_event))
            p.start()
            processes.append(p)

        # collect MOPG results for offsprings and insert objs into objs buffer
        all_offspring_batch = [[] for _ in range(len(processes))]
        cnt_done_workers = 0
        while cnt_done_workers < len(processes):
            rl_results = results_queue.get()
            task_id, offsprings = rl_results['task_id'], rl_results['offspring_batch']
            for sample in offsprings:
                all_offspring_batch[task_id].append(Sample.copy_from(sample))
            if rl_results['done']:
                cnt_done_workers += 1

        # put all intermidiate policies into all_sample_batch for EP update
        # all_sample_batch = [] 
        # store the last policy for each optimization weight for RA
        last_offspring_batch = [None] * len(processes) 
        # only the policies with iteration % update_iter = 0 are inserted into offspring_batch for population update
        # after warm-up stage, it's equivalent to the last_offspring_batch
        offspring_batch = [] 
        for task_id in range(len(processes)):
            offsprings = all_offspring_batch[task_id]
            #prev_node_id = task_batch[task_id].sample.optgraph_id
            #opt_weights = deepcopy(task_batch[task_id].scalarization.weights).detach().numpy()
            for i, sample in enumerate(offsprings):
                all_sample_batch.append(sample)
                if (i + 1) % args.update_iter == 0:
                    #prev_node_id = opt_graph.insert(opt_weights, deepcopy(sample.objs), prev_node_id)
                    #sample.optgraph_id = prev_node_id
                    offspring_batch.append(sample)
            last_offspring_batch[task_id] = offsprings[-1]

        done_event.set()

        # -----------------------> Update EP <----------------------- #
        # update EP and population
        print('update EP and population')
        ep.update(all_sample_batch)

        # ------------------- > Task Selection <--------------------- #
        elite_batch = ep.sample_batch
        ep_objs = []
        print_info('Selected Tasks:')
        for i in range(len(elite_batch)):
            print_info('objs = {}'.format(elite_batch[i].objs))
            ep_objs.append(elite_batch[i].objs)
            print(len(all_sample_batch))
        ep_objs_array = np.array(ep_objs)
        torch.save(ep_objs_array,os.path.join(args.save_dir, 'ep_objs', 'extension_iter_'+str(extension_iter+1)+'.pt'))

        # save
        reward_list = []
        for sample in all_sample_batch:
            reward_list.append(sample.objs)
        reward_list_array = np.array(reward_list)
        torch.save(reward_list_array,os.path.join(args.save_dir, 'objs', 'extension_iter_'+str(extension_iter+1)+'.pt'))
        hv = eval(reward_list_array, ref)
        #wandb.log({"hv": hv},step = extension_iter)
    end_time = time.time()
    print('total time:', end_time - start_time)